from .slide import Slide as slide
from .patcher import Patcher as patcher
from .annotation import Annotation as annotation
from .rule import Rule as rule
from .converter import Converter as converter

__version__ = "1.1.1"
