from .ASAP_parser import ASAPAnnotation
from .QuPath_parser import QuPathAnnotation
from .SlideRunner_parser import SlideRunnerAnnotation
from .WSIDissector_parser import WSIDissectorAnnotation
from .NDPView_parser import NDPViewAnnotation
from .parser_utils import BaseParser
